/*
 * tamba_model_tx_intf_enet.v
 *
 * Description:
 *	Tamba Tx Interface Ethernet Stream model
 *
 */

module tamba_model_tx_intf_enet #
  (
     /* ext
      */
    parameter SBC  = 64,
    parameter PWC  = 1,
    parameter BPW  = 8,
   
     /* rt
      */
    parameter RT  = 1,

     /* log2
      */
    parameter log2SBC  = ((SBC>256)?9:
			  (SBC>128)?8:
			  (SBC> 64)?7:
			  (SBC> 32)?6:
			  (SBC> 16)?5:
			  (SBC>  8)?4:
			  (SBC>  4)?3:
			  (SBC>  2)?2:1),
   
     /* int
      */
    parameter PBC     = PWC*BPW,
    parameter bitPW   = BPW*8,

    parameter ST_SOP  = 1'b0,
    parameter ST_PKT  = 1'b1
   )
  (
    input                 i_clk,
    input 		  i_sreset,

     /* cfg
      */
    input [15:0] 	  i_cfg_size_min,
    input [15:0] 	  i_cfg_size_max,
    input [15:0] 	  i_cfg_rep,
    input [7:0] 	  i_cfg_ipg,
    input 		  i_cfg_pad,
    input [7:0] 	  i_cfg_pad_size,
    input 		  i_cfg_crc,
   
     /* control
      */
    input 		  i_start_strb,
    output 		  o_first_strb,		  
    output 		  o_end_strb,

     /* Stream Interface
      */
    input 		  i_fc,
   
    output 		  o_en,
    output 		  o_sop,
    output 		  o_eop,
    output 		  o_eope,		 
    output [7:0] 	  o_bc,
    output [PBC*8-1:0] 	  o_preamble,
    output [SBC*8-1:0] 	  o_data
   ); 

/*************************************************************************************
 * Register declarations 
 * 
 *************************************************************************************/
   reg 			 start_n;
   
   reg 			 on_n;
   reg 			 on_r;

   reg [15:0] 		 rep_n;
   reg [15:0] 		 rep_r;

   reg 			 pkt_en_n;
   reg [15:0] 		 pkt_size_n;
   reg [15:0] 		 pkt_size_r;
   reg [15:0] 		 pkt_size_wpad_n;
   reg [15:0] 		 pkt_size_wcrc_n;
   
   reg [7:0] 		 ifg_cnt_n;
   reg [7:0] 		 ifg_cnt_r;
   reg 			 ifg_on_n;
   
   reg [2048*8-1:0] 	 data_n;
      
   reg 			 d;
   reg [31:0] 		 c;
   reg [31:0] 		 crc_n;
   reg [31:0] 		 crc_n_z;
      

   reg 			 en_n;
   reg [15:0] 		 st_size_n;
   reg [15:0] 		 st_size_r;
   reg 			 st_n;
   reg 			 st_r;
   
   reg [BPW*8*8-1:0] 	 preamble_n;
   reg [BPW*8*8-1:0] 	 preamble_r;

   reg 			 tx_en_n;
   reg 			 tx_sop_n;
   reg 			 tx_eop_n;
   reg [log2SBC-1:0] 	 tx_bc_n;
   reg [SBC*8-1:0] 	 tx_data_n;
   reg [SBC*8-1:0] 	 tx_data_z_n;
   reg [PBC*8-1:0] 	 tx_preamble_n;
         
   reg 			 tx_en_r;
   reg 			 tx_eop_r;
   
   reg 			 tx_out_en_n;
   reg 			 tx_out_sop_n;
   reg 			 tx_out_eop_n;
   reg 			 tx_out_eope_n;
   reg [log2SBC-1:0] 	 tx_out_bc_n;
   reg [SBC*8-1:0] 	 tx_out_data_n;
   reg [PBC*8-1:0] 	 tx_out_preamble_n;
   
   reg 			 tx_out_en_r;
   reg 			 tx_out_sop_r;
   reg 			 tx_out_eop_r;
   reg 			 tx_out_eope_r;
   reg [log2SBC-1:0] 	 tx_out_bc_r;
   reg [SBC*8-1:0] 	 tx_out_data_r;
   reg [PBC*8-1:0] 	 tx_out_preamble_r;

     /*
      */
   integer 		 i,j;
   
/*************************************************************************************
 * Always, start_
 *
 *************************************************************************************/
   always @(*)
     start_n  = (i_start_strb /*& !on_r*/);
      
/*************************************************************************************
 * Always, on_
 *
 *************************************************************************************/
   always @(*)
     begin
	on_n  = on_r;
	if (start_n)
	  on_n  = 1'b1;
	else if (tx_en_r & tx_eop_r)
	  begin
	     if (rep_r == i_cfg_rep)
	       begin
		  if (pkt_size_r == i_cfg_size_max)
		    on_n  = 1'b0;
	       end
	  end
     end
   
   always @(posedge i_clk)
     if (i_sreset)
       on_r  <= 1'b0;
     else
       on_r  <= on_n;
   
/*************************************************************************************
 * Always, rep_
 *
 *************************************************************************************/
   always @(*)
     begin
	rep_n  = rep_r;
	if (start_n)
	  rep_n  = 1;
	else if (tx_en_r & tx_eop_r)
	  begin
	     if (rep_r == i_cfg_rep)
	       rep_n  = 1;
	     else
	       rep_n  = (rep_r + 1);
	  end
     end

   always @(posedge i_clk)
     rep_r  <= rep_n;
		  
/*************************************************************************************
 * Always, pkt_en_
 *
 *************************************************************************************/
   always @(*)
     pkt_en_n  = (on_n & (start_n | (tx_en_r & tx_eop_r)));

/*************************************************************************************
 * Always, pkt_size_
 *
 *************************************************************************************/
   always @(*)
     begin
	pkt_size_n  = pkt_size_r;
	if (start_n)
	  pkt_size_n  = i_cfg_size_min;
	else if (tx_en_r & tx_eop_r)
	  begin
	     if (rep_r == i_cfg_rep)
	       pkt_size_n  = (pkt_size_r + 1);
	  end
     end

   always @(posedge i_clk)
     pkt_size_r  <= pkt_size_n;
   
   always @(*)
     pkt_size_wpad_n  = ((i_cfg_pad & (pkt_size_n < (i_cfg_pad_size - 4))) ? (i_cfg_pad_size - 4) : pkt_size_n);
   
   always @(*)
     pkt_size_wcrc_n  = (i_cfg_crc ? (pkt_size_wpad_n + 4) : pkt_size_wpad_n);

/*************************************************************************************
 * Always, ifg_
 *
 *************************************************************************************/
   always @(*)
     begin
	if (start_n)
	  ifg_cnt_n  = 0;
	else if (tx_en_r & tx_eop_r)
	  ifg_cnt_n  = i_cfg_ipg;
	else if (ifg_cnt_r != 0)
	  ifg_cnt_n  = (ifg_cnt_r - 1'h1);
	else
	  ifg_cnt_n  = ifg_cnt_r;
     end
	
   always @(posedge i_clk)
     ifg_cnt_r  <= ifg_cnt_n;

   always @(*)
     ifg_on_n  = (ifg_cnt_n != 0);
   
/*************************************************************************************
 * Always, data_
 *
 *************************************************************************************/
   always @(*)
     for (i=0;i<(i_cfg_pad_size-4);i=i+1)
       begin
	  if (i >= pkt_size_n)
	    data_n[i*8+:8]  = 8'h00;
	  else
	    data_n[i*8+:8] = i[7:0]+1;
       end

   initial
     for (i=0;i<2048;i=i+1)
       data_n[i*8+:8] = i[7:0]+1;
     
/*************************************************************************************
 * Always, crc_
 *
 *************************************************************************************/
   always @(*)
     begin
	for (i=0;i<2048*8;i=i+1)
	  begin
	     if (i<(pkt_size_wpad_n*8))
	       begin
		  d  = ((i<32) ? !data_n[i] : data_n[i]);
		  c  = ((i==0) ? 32'h0 : crc_n);
		  
		  crc_n[0]  = d ^ c[31];
		  crc_n[1]  = d ^ c[0] ^ c[31];
		  crc_n[2]  = d ^ c[1] ^ c[31];
		  crc_n[3]  = c[2];
		  crc_n[4]  = d ^ c[3] ^ c[31];
		  crc_n[5]  = d ^ c[4] ^ c[31];
		  crc_n[6]  = c[5];
		  crc_n[7]  = d ^ c[6] ^ c[31];
		  crc_n[8]  = d ^ c[7] ^ c[31];
		  crc_n[9]  = c[8];
		  crc_n[10] = d ^ c[9] ^ c[31];
		  crc_n[11] = d ^ c[10] ^ c[31];
		  crc_n[12] = d ^ c[11] ^ c[31];
		  crc_n[13] = c[12];
		  crc_n[14] = c[13];
		  crc_n[15] = c[14];
		  crc_n[16] = d ^ c[15] ^ c[31];
		  crc_n[17] = c[16];
		  crc_n[18] = c[17];
		  crc_n[19] = c[18];
		  crc_n[20] = c[19];
		  crc_n[21] = c[20];
		  crc_n[22] = d ^ c[21] ^ c[31];
		  crc_n[23] = d ^ c[22] ^ c[31];
		  crc_n[24] = c[23];
		  crc_n[25] = c[24];
		  crc_n[26] = d ^ c[25] ^ c[31];
		  crc_n[27] = c[26];
		  crc_n[28] = c[27];
		  crc_n[29] = c[28];
		  crc_n[30] = c[29];
		  crc_n[31] = c[30];
	       end
	  end
     end
	
   always @(*)
     for (i=0;i<4;i=i+1)
       for (j=0;j<8;j=j+1)
	 crc_n_z[i*8+j+:1]  = crc_n[i*8+(7-j)+:1];
	
/*************************************************************************************
 * Always, en_
 *
 *************************************************************************************/
   always @(*)
     en_n  = (on_n & !i_fc & !ifg_on_n);
   
/*************************************************************************************
 * Always, st_
 *
 *************************************************************************************/
   always @(*)
     begin
	st_size_n  = st_size_r;
	st_n       = st_r;

	if (start_n & 1'b0)
	  begin
	     st_size_n  = 0;
	     if (pkt_size_wcrc_n > SBC)
	       st_n  = ST_PKT;
	  end
	else if (en_n)
	  begin
	     case (st_r)
	       ST_SOP:
		 begin
		    st_size_n  = 0;
		    if (pkt_size_wcrc_n > SBC)
		      st_n  = ST_PKT;
		 end
	       ST_PKT:
		 begin
		    st_size_n  = (st_size_r + SBC);
		    if ((pkt_size_wcrc_n - st_size_n) <= SBC)
		      st_n  = ST_SOP;
		 end
	     endcase
	  end
     end 

   always @(posedge i_clk)
     if (i_sreset)
       st_r  <= ST_SOP;
     else
       st_r  <= st_n;
       
   always @(posedge i_clk)
     st_size_r  <= st_size_n;

/*************************************************************************************
 * Always, preamble_
 *
 *************************************************************************************/
   always @(*)
     begin
	preamble_n         = preamble_r;
	preamble_n[8+:16]  = pkt_size_n;

	for (i=0;i<8;i=i+1)
	  preamble_n[i*bitPW+8+:16]  = pkt_size_n;
     end

   always @(posedge i_clk)
     if (i_sreset)
       begin
	  for (i=0;i<8;i=i+1)
	    for (j=0;j<BPW;j=j+1)
	      preamble_r[i*bitPW+j*8+:8]  = (i+1);
       end
     else
       preamble_r  <= preamble_n;
      
/*************************************************************************************
 * Always, tx_
 *
 *************************************************************************************/
   always @(*)
     begin
	if (en_n)
	  begin
	     tx_en_n   = 1'b1;

	     tx_sop_n  = ((st_r == ST_SOP) | start_n);
	     tx_eop_n  =  (st_n == ST_SOP);

	     tx_bc_n   = (((pkt_size_wcrc_n - st_size_n) < SBC) ? (pkt_size_wcrc_n-1) : {log2SBC{1'b1}});
	     
	     for (i=0;i<SBC;i=i+1)
	       if ((pkt_size_wcrc_n - (st_size_n+1)) < i)
		 tx_data_z_n[i*8+:8]  = 8'hxx;
	       else
		 if (i_cfg_crc)
		   case (pkt_size_wcrc_n - (st_size_n+i))
		     4:        tx_data_z_n[i*8+:8]  = ~crc_n_z[31:24];
		     3:        tx_data_z_n[i*8+:8]  = ~crc_n_z[23:16];
		     2:        tx_data_z_n[i*8+:8]  = ~crc_n_z[15:8];
		     1:        tx_data_z_n[i*8+:8]  = ~crc_n_z[7:0];
		     default:  tx_data_z_n[i*8+:8]  = data_n[(st_size_n+i)*8+:8];
		   endcase
		 else
		   tx_data_z_n[i*8+:8]  = data_n[(st_size_n+i)*8+:8];

	     tx_preamble_n  = preamble_n[0+:PBC*8];
	  end
	else
	  begin
	     tx_en_n        = 1'b0;
	     tx_sop_n       = 1'bx;
	     tx_eop_n       = 1'bx;
	     tx_bc_n        = {log2SBC{1'bx}};
	     tx_data_z_n    = {SBC{8'hxx}};
	     tx_preamble_n  = {PBC{8'hxx}};
	  end
     end
		
   always @(*)
     for (i=0;i<SBC;i=i+1)
       tx_data_n[i*8+:8]  = tx_data_z_n[(SBC-1-i)*8+:8];

/*************************************************************************************
 * Always, tx_
 *
 *************************************************************************************/
   always @(posedge i_clk)
     begin
	tx_en_r   <= tx_en_n;
	tx_eop_r  <= tx_eop_n;
     end

   always @(*)
     begin
	tx_out_en_n        =                    tx_en_n;
	tx_out_sop_n       =                    tx_sop_n;
	tx_out_eop_n       =                    tx_eop_n;
	tx_out_eope_n      =                   (tx_eop_n & (pkt_size_n < 14) & i_cfg_pad);
	tx_out_bc_n        = ((SBC==1) ? 1'b0 : tx_bc_n);
	tx_out_data_n      =                    tx_data_n;
	tx_out_preamble_n  =                    tx_preamble_n;
     end

   always @(posedge i_clk)
     begin
	tx_out_en_r        <= tx_out_en_n;
	tx_out_sop_r       <= tx_out_sop_n;
	tx_out_eop_r       <= tx_out_eop_n;
	tx_out_eope_r      <= tx_out_eope_n;
	tx_out_bc_r        <= tx_out_bc_n;
	tx_out_data_r      <= tx_out_data_n;
	tx_out_preamble_r  <= tx_out_preamble_n;
     end

/*************************************************************************************
 * Always, first_
 *
 *************************************************************************************/
   reg first_r;
   reg first_strb_n;
   
   always @(posedge i_clk)
     if (o_en)
       first_r  <= 1'b0;
     else if (i_sreset | ( on_n & !on_r))
       first_r  <= 1'b1;
      
   always @(*)
     first_strb_n  = (o_en & (first_r | (on_n & !on_r)));
   
/*************************************************************************************
 * Assignment
 *
 *************************************************************************************/
     /* ctrl
      */
   assign o_first_strb  = first_strb_n;
   assign o_end_strb    = (!on_n &  on_r);

     /* stream
      */
   assign o_en        = (RT ? tx_out_en_r       : tx_out_en_n);
   assign o_sop       = (RT ? tx_out_sop_r      : tx_out_sop_n);
   assign o_eop       = (RT ? tx_out_eop_r      : tx_out_eop_n);
   assign o_eope      = (RT ? tx_out_eope_r     : tx_out_eope_n);
   assign o_bc        = (RT ? tx_out_bc_r       : tx_out_bc_n);
   assign o_preamble  = (RT ? tx_out_preamble_r : tx_out_preamble_n);
   assign o_data      = (RT ? tx_out_data_r     : tx_out_data_n);

endmodule

